from pymongo import MongoClient
from os import environ
import logging, requests, time
from ghost_ship import exception

MONGO_URL = environ['TYK_MONGO_URL']
CONSUL_ADDR = environ['CONSUL_ADDR']
TYK_ADDR = environ['TYK_ADDR']
TYK_TOKEN = environ['TYK_TOKEN']
TYK_DB_NAME = environ.get('TYK_DB_NAME', 'tyk_analytics')

mong = MongoClient(MONGO_URL)
db = mong[TYK_DB_NAME]

def reload(service_name, state):
  logging.info('Waiting 10s before reload...')
  query_endpoint = CONSUL_ADDR + '/v1/health/service/%s?passing&tag=%s' % (service_name, state)
  condition = {"proxy.service_discovery.query_endpoint": {"$regex": "http(.)*%s(.)*" % service_name}}
  logging.debug('Tyk update condition')
  logging.debug(condition)
  for item in db.tyk_apis.find(condition):
    logging.debug(item)

  if db.tyk_apis.find(condition).count():
    logging.info('Update %s discovery to: %s' % (service_name, query_endpoint))
    db.tyk_apis.update_many(condition, {
      "$set": {
        "proxy.service_discovery.query_endpoint": query_endpoint 
      }
    })
    logging.info('Reload tyk')
    if not reload_tyk():
      raise exception.TykReloadFailed('Reload Tyk fail cmnr!!!')
    logging.info('Reload signal has been sent to tyk, waiting 30s...')
    time.sleep(30)
  else:
    raise exception.TykServiceNotFound('Tyk service: %s is not found' % service_name)

def reload_tyk():
  url = TYK_ADDR + '/tyk/reload/group'
  headers = {
    "X-Tyk-Authorization": TYK_TOKEN
  }
  resp = requests.get(url, headers=headers)
  return resp.status_code == 200



from jinja2 import Environment, PackageLoader, FileSystemLoader
import nomad, os, exception, loader, time, tyk
from urlparse import urlparse
import logging

FORMAT = '[GHOST SHIP] - [%(levelname)s]\t - %(message)s'

loglevel = os.environ.get('GHOSTSHIP_LOGLEVEL', 'info')
level = getattr(logging, loglevel.upper(), None)
logging.basicConfig(format=FORMAT, level=level)

NOMAD_ADDR = os.environ['NOMAD_ADDR']
nomad_url_object = urlparse(NOMAD_ADDR) 

nomad_client = nomad.Nomad(nomad_url_object.hostname,timeout=5)
# base_env for jobs and groups templates
base_env = Environment(loader=PackageLoader('ghost_ship', 'templates'))
# context env for jobs
logging.info('Current directory: %s' % os.getcwd())

cont_env = Environment(loader=FileSystemLoader(os.getcwd() + '/templates'))

class Job:
  def __init__(self, name):
    self.name = name
    self.groups = {
      'green': self.default_for('green'),
      'blue': self.default_for('blue')
    }

  def default_for(self, group_name):
    return {
      'service_name': self.name,
      'version': 'latest',
      'count': 0,
      'group': group_name
    }

  def run(self):
    print loader.run(loader.write(self.name, self.render()))

  def initilize(self, version):
    self.groups['green']['version'] = version
    del self.groups['green']['count']
    self.run()
    logging.info('Initilized with version %s' % version)
    try:
      tyk.reload(self.name, 'green')
    except exception.TykServiceNotFound:
      logging.info('Missing tyk service, tyk need manual config')


  def upgrade(self, version='0'):
    self.inspect()
    next = 'green' if self.current == 'blue' else 'blue'

    logging.info('Current state: %s' % self.current)
    logging.info('Next state: %s' % next)

    logging.info('Current version: %s' % self.groups[self.current]['version'])
    logging.info('Do upgrade to: %s' % version)

    self.groups[next]['count'] = self.groups[self.current]['count']
    self.groups[next]['version'] = version
    
    logging.debug(self.groups)
    self.run()
    if self.waiting(next):
      tyk.reload(self.name, next) 
      logging.info('Adjust count of %s to 0' % self.current)
      self.groups[self.current]['count'] = 0
      self.run()
    else:
      logging.error('Upgrade failed, rollback to %s' % self.current)
      self.groups[next]['count'] = 0
      self.run()
      exit(1)

  def scale(self, count=1):
    self.inspect() 
    logging.info('Scale %s to %s' % (self.current, count))
    self.groups[self.current]['count'] = count
    self.run()
    self.waiting(self.current)
    logging.info('What is done is done')

  def rollback(self):
    pass

  def inspect(self):
    try:
      job_info = nomad_client.job[self.name]
    except:
      raise exception.JobNotFound(self.name)

    for group in job_info['TaskGroups']:
      group_name = group['Name']
      instances_count = group['Count']

      image_version = (group['Tasks'][0]['Config']['image']).split(':')[1]

      self.groups[group_name] = {
        'service_name': self.name,
        'count': instances_count,
        'version': image_version,
        'group': group_name
      }

    self.current = 'green' if self.groups['green']['count'] >= self.groups['blue']['count'] else 'blue'

  def render_context(self, name):
    try:
      return cont_env.get_template(self.name + '.nomad.jinja').render(**self.groups.get(name, self.default_for('name')))
    except:
      raise exception.JobTemplateIsNotFound(self.name)

  def render_group(self, name):
    return base_env.get_template('group.nomad.jinja').render(group_name=name, content=self.render_context(name))

  def render(self):
    data = {
      'green': self.render_group('green'),
      'blue': self.render_group('blue'),
      'job_name': self.name
    }
    return base_env.get_template('job.nomad.jinja').render(**data)

  def waiting(self, group_name):
    running = True
    logging.info('Starting containers...', dict(continued=True))
    while (running):
      logging.info('...')
      j = nomad_client.jobs[self.name]
      logging.debug(j['JobSummary']['Summary'][group_name])
      running = j['JobSummary']['Summary'][group_name]['Starting'] != 0
      time.sleep(5)

    return j['JobSummary']['Summary'][group_name]['Running'] == self.groups[group_name]['count']



import subprocess

def write(name, content):
  filename = '/tmp/' + name + '.nomad'
  f = open(filename, 'w')
  f.write(content)
  f.close()
  return filename

def run(filename):
  return subprocess.check_output(['nomad', 'run', filename])  



class JobNotFound(Exception):
  pass

class GroupNotFound(Exception):
  pass

class UpgradeFail(Exception):
  pass

class TykServiceNotFound(Exception):
  pass

class TykReloadFailed(Exception):
  pass
class JobTemplateIsNotFound(Exception):
  pass



#!/usr/bin/env python

from os import environ
from sys import argv

NOMAD_ADDR = environ['NOMAD_ADDR']
MONGO_URL = environ['TYK_MONGO_URL']
CONSUL_ADDR = environ['CONSUL_ADDR']
TYK_ADDR = environ['TYK_ADDR']
TYK_TOKEN = environ['TYK_TOKEN']

VERSION = '0.2.5'

from ghost_ship.job import Job
from ghost_ship.exception import *
from ghost_ship import tyk
from getopt import getopt
import logging 

usage = """Usage:

* Init a service

  $ ghost-ship run service-name --scale=4

* scale service to 2 containers

  $ ghost-ship scale service-name 2
 
* Upgrade service to version v2

  $ ghost-ship upgrade service-name v2
 
"""


def run(service_name, version):
  Job(service_name).initilize(version)

def scale(service_name, count):
  try:
    Job(service_name).scale(int(count))
  except:
    logging.error('Count must be int')
    exit(1)

def upgrade(service_name, version):
  Job(service_name).upgrade(version)

def version():
  print 'Ghost ship v%s' % VERSION

job_commands = dict(run=run, upgrade=upgrade, scale=scale, version=version)

def do(command, args):
  if command != 'version':
    info() 
  job_commands[command](*args)  

def info():
  logging.info('NOMAD_ADDR: %s ' %NOMAD_ADDR)
  logging.info('TYK_MONGO_URL: %s ' % MONGO_URL)
  logging.info('CONSUL_ADDR: %s ' % CONSUL_ADDR)
  logging.info('TYK_ADDR: %s ' % TYK_ADDR)
  logging.info('TYK_TOKEN: %s ' % TYK_TOKEN)

def main():
  try:
    if len(argv) >= 2:
      do(argv[1], argv[2:])
    else:
      print usage
  except JobNotFound as e:
    logging.error('Job not found: %s', e)
    exit(1)
  except GroupNotFound as e:
    logging.error('Job Group is not found: %s', e)
    exit(1)
  except UpgradeFail as e:
    logging.error('Job upgrade fail: %s', e)
    exit(1)
  except TykServiceNotFound as e:
    logging.error('Tyk service is not found: %s', e)
    exit(1)
  except TykReloadFailed as e:
    logging.error('Tyk reload was failed: %s', e)
    exit(1)
  except JobTemplateIsNotFound as e:
    logging.error('Job template is not found: %s', e)
    exit(1)

if __name__ == '__main__':
  main()



#!/usr/bin/env python

from distutils.core import setup

setup(
  name='ghost-ship',
  version='0.3.5',
  description='Nomad ghost ship deploy',
  author='Thinh Tran',
  author_email='duythinht',
  url='http://git.chotot.org/corex/ghost-ship',
  packages=['ghost_ship'],
  package_data={'ghost_ship': ['templates/*.jinja']},
  py_modules=['pirate'],
  install_requires=['requests', 'python-nomad', 'jinja2', 'pymongo'],
  scripts=['scripts/ghost-ship']
)
